#include "hardware.h"
#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include "utils.h"
#include <util/delay.h>
#include "isp.h"
#include "timer.h"
#include "oddebug.h"

namespace stk{

static uchar    ispClockDelay;
static uchar    cmdBuffer[4];

/* ------------------------------------------------------------------------- */
/* We disable interrupts while transfer a byte. This ensures that we execute
 * at nominal speed.
 */
static byte ispBlockTransfer(const byte*block, uchar len) {
 byte cnt, shift = 0, port, delay = ispClockDelay;

/* minimum clock pulse width:
 * 5 + 4 * delay clock cycles           -> Tmin = 750 ns
 * total clock period: 12 + 8 * delay   -> fmax = 600 kHz
 */
/*    DBG2(0x40, block, len); */
    cli();
    port = PORT_OUT(HWPIN_ISP_MOSI) & ~(1 << PORT_BIT(HWPIN_ISP_MOSI));
    while(len--){   /* len may be 0 */
        cnt = 8;
        shift = *block++;
        do{
            if(shift & 0x80){
                port |= (1 << PORT_BIT(HWPIN_ISP_MOSI));
            }
            PORT_OUT(HWPIN_ISP_MOSI) = port;
            sei();
            timerTicksDelay(delay);
            cli();
            PORT_PIN_SET(HWPIN_ISP_SCK);    /* <-- data clocked by device */
            shift <<= 1;
            port &= ~(1 << PORT_BIT(HWPIN_ISP_MOSI));
            if(PORT_PIN_VALUE(HWPIN_ISP_MISO))  /* no driver in this hardware */
                shift |= 1;
            sei();
            timerTicksDelay(delay);
            cli();
            PORT_PIN_CLR(HWPIN_ISP_SCK);    /* <-- device changes data */
        }while(--cnt);
    }
    sei();
/*    DBG2(0x41, &shift, 1); */
    return shift;
}

/* ------------------------------------------------------------------------- */

static void ispAttachToDevice(uchar stk500Delay, uchar stabDelay)
{
#if !METABOARD_HARDWARE /* on metaboard, we use the jumper to select CDC vs. HID mode */
    if(!PORT_PIN_VALUE(HWPIN_JUMPER)){      /* Jumper is set -> request clock below 8 kHz */
        ispClockDelay = (uchar)(70/TIMER_TICK_US);   /* 140 us -> 7.14 kHz clock rate */
    }else
#endif
    if(stk500Delay == 0){ /* 1.8 MHz nominal */
        ispClockDelay = 0;
    }else if(stk500Delay == 1){ /* 460 kHz nominal */
        ispClockDelay = 0;
    }else if(stk500Delay == 2){ /* 115 kHz nominal */
        ispClockDelay = 1;
    }else if(stk500Delay == 3){ /* 58 kHz nominal */
        ispClockDelay = 2;
    }else{
        /* from STK500v2 spec: stk500Delay = 1/(24 * SCK / 7.37 MHz) - 10/12
         * definition of ispClockDelay = 1 + 1/(SCK/MHz * 2 * TIMER_TICK_US)
         * ispClockDelay = 1 + (stk500Delay + 10/12) * 12 / (TIMER_TICK_US * 7.37)
         */
#if F_CPU > 14000000L    /* ~ 16 MHz */
        ispClockDelay = (stk500Delay + 1)/2 - (stk500Delay + 1)/8 + (stk500Delay + 1)/32;
#else   /* must be 12 MHz */
        ispClockDelay = (stk500Delay + 1)/4 + (stk500Delay + 1)/16;
#endif
    }
#if METABOARD_HARDWARE
    PORT_PIN_CLR(HWPIN_LED);        /* turn on LED */
    /* turn on power for programming sockets: */
    PORT_OUT(HWPIN_ISP_SUPPLY1) |= _BV(PORT_BIT(HWPIN_ISP_SUPPLY1)) | _BV(PORT_BIT(HWPIN_ISP_SUPPLY2));
#ifdef HWPIN_ISP_CLK
    PORT_DDR_SET(HWPIN_ISP_CLK);    /* turn on optional clock for programming socket */
#endif
    PORT_DDR_SET(HWPIN_ISP_SCK);    /* enable programming I/O pins */
    PORT_DDR_SET(HWPIN_ISP_MOSI);
    timerMsDelay(200);              /* allow device to power up */
    PORT_DDR_SET(HWPIN_ISP_RESET);
    timerMsDelay(stabDelay);
    timerTicksDelay(ispClockDelay); /* stabDelay may have been 0 */
    PORT_PIN_SET(HWPIN_ISP_RESET);  /* give positive RESET pulse */
    timerTicksDelay(ispClockDelay);
    PORT_PIN_CLR(HWPIN_ISP_RESET);  /* bring RESET back low */
#else /* METABOARD_HARDWARE */
    PORT_PIN_SET(HWPIN_LED);
    /* setup initial condition: SCK, MOSI = 0 */
    PORT_OUT(HWPIN_ISP_SCK) &= ~((1 << PORT_BIT(HWPIN_ISP_SCK)) | (1 << PORT_BIT(HWPIN_ISP_MOSI)));
    PORT_PIN_SET(HWPIN_ISP_RESET);  /* set RESET */
    PORT_DDR_CLR(HWPIN_ISP_DRIVER); /* make input: use internal pullup to control driver */
    PORT_PIN_SET(HWPIN_ISP_DRIVER); /* attach to device: */
//    TCCR2 |= (1 << COM20);  /* set toggle on compare match mode -> activate clock */
    timerMsDelay(stabDelay);
    timerTicksDelay(ispClockDelay); /* stabDelay may have been 0 */
    /* We now need to give a positive pulse on RESET since we can't guarantee
     * that SCK was low during power up (according to instructions in Atmel's
     * data sheets).
     */
    PORT_PIN_CLR(HWPIN_ISP_RESET);  /* give a positive RESET pulse */
    timerTicksDelay(ispClockDelay);
    PORT_PIN_SET(HWPIN_ISP_RESET);
#endif /* METABOARD_HARDWARE */
}

static void ispDetachFromDevice(uchar removeResetDelay)
{
#if METABOARD_HARDWARE
    PORT_DDR_CLR(HWPIN_ISP_SCK);
    PORT_DDR_CLR(HWPIN_ISP_MOSI);
#ifdef HWPIN_ISP_CLK
    PORT_DDR_CLR(HWPIN_ISP_CLK);    /* turn off optional clock for programming socket */
#endif
    PORT_DDR_CLR(HWPIN_ISP_RESET);
    PORT_PIN_CLR(HWPIN_ISP_SCK);    /* ensure we have no pull-ups active */
    PORT_PIN_CLR(HWPIN_ISP_MOSI);
    /* turn off optional power supply for programming socket: */
    PORT_OUT(HWPIN_ISP_SUPPLY1) &= ~(_BV(PORT_BIT(HWPIN_ISP_SUPPLY1)) | _BV(PORT_BIT(HWPIN_ISP_SUPPLY2)));
    PORT_PIN_SET(HWPIN_LED);        /* turn off LED */
#else /* METABOARD_HARDWARE */
    PORT_OUT(HWPIN_ISP_SCK) &= ~((1 << PORT_BIT(HWPIN_ISP_SCK)) | (1 << PORT_BIT(HWPIN_ISP_MOSI)));
    PORT_PIN_SET(HWPIN_ISP_RESET);
    timerMsDelay(removeResetDelay);
//    TCCR2 &= ~(1 << COM20);  /* clear toggle on compare match mode */
    PORT_PIN_CLR(HWPIN_ISP_DRIVER); /* detach from device */
    PORT_DDR_SET(HWPIN_ISP_DRIVER); /* set pin level to low-Z 0 */
    PORT_PIN_CLR(HWPIN_ISP_RESET);
    PORT_PIN_CLR(HWPIN_LED);
#endif /* METABOARD_HARDWARE */
}

/* ------------------------------------------------------------------------- */

byte isp::EnterProgmode() {
 auto&p=gParam.ispEnterProg;
 byte rval;

 ispAttachToDevice(stkParam.sckDuration, p.stabDelay);
 timerMsDelay(p.cmdExeDelay);
    /* we want for(i = param->synchLoops; i--;), but avrdude sends synchLoops == 0 */
 for(byte i = 32; i--;){
        wdt_reset();
        rval = ispBlockTransfer(p.cmd, p.pollIndex);
        if(p.pollIndex < 4)
            ispBlockTransfer(p.cmd + p.pollIndex, 4 - p.pollIndex);
        if(rval == p.pollValue){   /* success: we are in sync */
            return STK_STATUS_CMD_OK;
        }
        /* insert one clock pulse and try again: */
        PORT_PIN_SET(HWPIN_ISP_SCK);
        timerTicksDelay(ispClockDelay);
        PORT_PIN_CLR(HWPIN_ISP_SCK);
        timerTicksDelay(ispClockDelay);
    }
 ispDetachFromDevice(0);
 return STK_STATUS_CMD_FAILED;   /* failure */
}

void isp::LeaveProgmode() {
 auto&p=gParam.ispLeaveProg;
 ispDetachFromDevice(p.preDelay);
 timerMsDelay(p.postDelay);
}

/* ------------------------------------------------------------------------- */

static bool deviceIsBusy() {
 cmdBuffer[0] = 0xf0;
 cmdBuffer[1] = 0;
 return ispBlockTransfer(cmdBuffer, 4) & 1;
}

static byte waitUntilReady(byte msTimeout) {
 timerSetupTimeout(msTimeout);
 while (deviceIsBusy()) if(timerTimeoutOccurred()) return STK_STATUS_RDY_BSY_TOUT;
 return STK_STATUS_CMD_OK;
}

/* ------------------------------------------------------------------------- */

byte isp::ChipErase() {
 auto&p=gParam.ispChipErase;
 byte maxDelay = p.eraseDelay,
      rval = STK_STATUS_CMD_OK;

 ispBlockTransfer(p.cmd, 4);
 if (p.pollMethod){
  if(maxDelay < 10) maxDelay = 10;  /* allow at least 10 ms */
  rval = waitUntilReady(maxDelay);
 }else timerMsDelay(maxDelay);
 return rval;
}

/* ------------------------------------------------------------------------- */

byte isp::ProgramMemory(bool isEeprom) {
 auto&p=gParam.ispProgramFlash;
 word numBytes = p.numBytes[0]<<8|p.numBytes[1];
 byte rval = STK_STATUS_CMD_OK;
uchar       valuePollingMask, rdyPollingMask;

 if (p.mode & 1){    /* page mode */
  valuePollingMask = 0x20;
  rdyPollingMask = 0x40;
 }else{                  /* word mode */
  valuePollingMask = 4;
  rdyPollingMask = 8;
 }
 if (!isEeprom && Address[3] & 0x80){
  cmdBuffer[0] = 0x4d;    /* load extended address */
  cmdBuffer[1] = 0x00;
  cmdBuffer[2] = Address[2];
  cmdBuffer[3] = 0x00;
  ispBlockTransfer(cmdBuffer, 4);
 }
 for(word i = 0; rval == STK_STATUS_CMD_OK && i < numBytes; i++){
  uchar x;
  wdt_reset();
  cmdBuffer[1] = Address[1];
  cmdBuffer[2] = Address[0];
  cmdBuffer[3] = p.data[i];
  x = p.cmd[0];
  if (!isEeprom){
   x &= ~0x08;
   if((uchar)i & 1){
    x |= 0x08;
    IncrementAddress();
   }
  }else IncrementAddress();
  cmdBuffer[0] = x;
  ispBlockTransfer(cmdBuffer, 4);
  if(p.mode & 1){            /* is page mode */
  if (i < numBytes - 1 || !(p.mode & 0x80))
                continue;               /* not last byte written */
  cmdBuffer[0] = p.cmd[1];     /* write program memory page */
  ispBlockTransfer(cmdBuffer, 4);
 }
        /* poll for ready after each byte (word mode) or page (page mode) */
 if(p.mode & valuePollingMask){ /* value polling */
  uchar d = p.data[i];
  if(d == p.poll[0] || (isEeprom && d == p.poll[1])){   /* must use timed polling */
                timerMsDelay(p.delay);
  }else{
                uchar x = p.cmd[2];     /* read memory */
                if(!isEeprom){
                    x &= ~0x08;
                    if((uchar)i & 1){
                        x |= 0x08;
                    }
                }
                cmdBuffer[0] = x;
                timerSetupTimeout(p.delay);
                while(ispBlockTransfer(cmdBuffer, 4) != d){
                    if(timerTimeoutOccurred()){
                        rval = STK_STATUS_CMD_TOUT;
                        break;
                    }
                }
            }
  }else if(p.mode & rdyPollingMask){ /* rdy/bsy polling */
            rval = waitUntilReady(p.delay);
  }else{                          /* must be timed delay */
            timerMsDelay(p.delay);
  }
 }
 return rval;
}

/* ------------------------------------------------------------------------- */

word isp::ReadMemory(bool isEeprom) {
 auto&pa = gParam.ispReadFlash;
 word numBytes = pa.numBytes[0]<<8|pa.numBytes[1];

    cmdBuffer[3] = 0;
    if(!isEeprom && Address[3] & 0x80){
        cmdBuffer[0] = 0x4d;    /* load extended address */
        cmdBuffer[1] = 0x00;
        cmdBuffer[2] = Address[2];
        ispBlockTransfer(cmdBuffer, 4);
    }
 byte*p = gResult.data;
 gResult.status = STK_STATUS_CMD_OK;
 byte cmd0 = pa.cmd;
 for(word i = 0; i < numBytes; i++){
        wdt_reset();
  cmdBuffer[1] = Address[1];
  cmdBuffer[2] = Address[0];
  if(!isEeprom){
            if((uchar)i & 1){
                cmd0 |= 0x08;
                IncrementAddress();
            }else{
                cmd0 &= ~0x08;
            }
        }else{
            IncrementAddress();
  }
  cmdBuffer[0] = cmd0;
  *p++ = ispBlockTransfer(cmdBuffer, 4);
 }
 *p = STK_STATUS_CMD_OK; /* status2 */
 return numBytes + 2;
}

/* ------------------------------------------------------------------------- */

byte isp::ProgramFuse() {
 ispBlockTransfer(gParam.ispProgramFuse.cmd, 4);
 return STK_STATUS_CMD_OK;
}

/* ------------------------------------------------------------------------- */

byte isp::ReadFuse() {
 auto&p=gParam.ispReadFuse;
 byte rval = ispBlockTransfer(p.cmd, p.retAddr);
 if(p.retAddr < 4)
        ispBlockTransfer(p.cmd + p.retAddr, 4 - p.retAddr);
 return rval;
}

/* ------------------------------------------------------------------------- */

uint isp::Multi() {
 auto&pa=gParam.ispMulti;
 byte cnt1, i, *p;

    cnt1 = pa.numTx;
    if(cnt1 > pa.rxStartAddr)
        cnt1 = pa.rxStartAddr;
    ispBlockTransfer(pa.txData, cnt1);

    p = gResult.data;
    for(i = 0; i < pa.numTx - cnt1; i++){
        uchar b = ispBlockTransfer(&pa.txData[cnt1] + i, 1);
        if(i < pa.numRx)
            *p++ = b;
        wdt_reset();
    }
    
    for(; i < pa.numRx; i++){
        cmdBuffer[0] = 0;
        *p++ = ispBlockTransfer(cmdBuffer, 1);
        wdt_reset();
    }
    *p = gResult.status = STK_STATUS_CMD_OK;
    return (uint)pa.numRx + 2;
}
}/*namespace stk*/